Package: alabaster.matrix
Title: Load and Save Artifacts from File
Version: 1.12.0
Date: 2025-03-04
Authors@R: person("Aaron", "Lun", role=c("aut", "cre"), email="infinite.monkeys.with.keyboards@gmail.com")
License: MIT + file LICENSE
Description: Save matrices, arrays and similar objects into file
        artifacts, and load them back into memory. This is a more
        portable alternative to serialization of such objects into RDS
        files. Each artifact is associated with metadata for further
        interpretation; downstream applications can enrich this
        metadata with context-specific properties.
Depends: alabaster.base
Imports: methods, BiocGenerics, S4Vectors, DelayedArray (>= 0.33.3),
        S4Arrays, SparseArray (>= 1.5.22), rhdf5 (>= 2.47.1),
        HDF5Array, Matrix, Rcpp
Suggests: testthat, knitr, BiocStyle, chihaya, BiocSingular,
        ResidualMatrix
LinkingTo: Rcpp
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Encoding: UTF-8
biocViews: DataImport, DataRepresentation
Config/pak/sysreqs: libssl-dev libnode-dev zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 13:00:04 UTC
RemoteUrl: https://github.com/bioc/alabaster.matrix
RemoteRef: RELEASE_3_23
RemoteSha: 66f650beff91c543d36e03d848655e15b0b07bd5
NeedsCompilation: yes
Packaged: 2026-04-29 13:48:19 UTC; root
Author: Aaron Lun [aut, cre]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 13:50:29 UTC; unix
