# Generated by roxygen2: do not edit by hand

export(.createRawArraySeed)
export(.extractArrayDimnames)
export(AmalgamatedArray)
export(AmalgamatedArraySeed)
export(DelayedMask)
export(ReloadedArray)
export(ReloadedArraySeed)
export(altReloadDelayedObject)
export(altReloadDelayedObjectFunction)
export(altStoreDelayedObject)
export(altStoreDelayedObjectFunction)
export(componentNames)
export(createExternalSeedDedupSession)
export(extractComponents)
export(loadArray)
export(loadWrapperArray)
export(preserveDelayedOperations)
export(readArray)
export(readDelayedArray)
export(readSparseMatrix)
export(recycleHdf5Files)
export(registerReloadDelayedObjectFunction)
export(reloadDelayedObject)
export(reloadDelayedObjectFunctionRegistry)
export(storeDelayedObject)
export(writeSparseMatrix)
exportClasses(AmalgamatedArray)
exportClasses(AmalgamatedArraySeed)
exportClasses(AmalgamatedMatrix)
exportClasses(DelayedMask)
exportClasses(ReloadedArray)
exportClasses(ReloadedArraySeed)
exportClasses(ReloadedMatrix)
exportClasses(WrapperArray)
exportClasses(WrapperArraySeed)
exportMethods(DelayedArray)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(extract_array)
exportMethods(extract_sparse_array)
exportMethods(is_sparse)
exportMethods(matrixClass)
exportMethods(path)
exportMethods(saveObject)
exportMethods(stageObject)
exportMethods(storeDelayedObject)
import(DelayedArray)
import(alabaster.base)
import(methods)
import(rhdf5)
importClassesFrom(DelayedArray,DelayedAbind)
importClassesFrom(DelayedArray,DelayedArray)
importClassesFrom(DelayedArray,DelayedMatrix)
importClassesFrom(DelayedArray,DelayedSetDimnames)
importClassesFrom(DelayedArray,DelayedUnaryOp)
importClassesFrom(Matrix,Matrix)
importClassesFrom(Matrix,dgCMatrix)
importClassesFrom(Matrix,dsparseMatrix)
importClassesFrom(SparseArray,SVT_SparseMatrix)
importFrom(BiocGenerics,"type<-")
importFrom(BiocGenerics,path)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,type)
importFrom(DelayedArray,DelayedArray)
importFrom(DelayedArray,RegularArrayGrid)
importFrom(DelayedArray,blockApply)
importFrom(DelayedArray,colAutoGrid)
importFrom(DelayedArray,currentViewport)
importFrom(DelayedArray,getAutoBlockLength)
importFrom(DelayedArray,getAutoBlockSize)
importFrom(DelayedArray,isPristine)
importFrom(DelayedArray,matrixClass)
importFrom(DelayedArray,new_DelayedArray)
importFrom(DelayedArray,rowAutoGrid)
importFrom(DelayedArray,seed)
importFrom(HDF5Array,H5SparseMatrixSeed)
importFrom(HDF5Array,HDF5Array)
importFrom(HDF5Array,HDF5ArraySeed)
importFrom(HDF5Array,getHDF5DumpChunkDim)
importFrom(HDF5Array,getHDF5DumpCompressionLevel)
importFrom(HDF5Array,writeHDF5Array)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(S4Arrays,acbind)
importFrom(S4Arrays,arbind)
importFrom(S4Arrays,extract_array)
importFrom(S4Arrays,is_sparse)
importFrom(S4Vectors,new2)
importFrom(SparseArray,extract_sparse_array)
importFrom(SparseArray,nzvals)
importFrom(SparseArray,nzwhich)
importFrom(SparseArray,read_block_as_sparse)
importFrom(alabaster.base,.loadObject)
importFrom(alabaster.base,.stageObject)
importFrom(alabaster.base,.writeMetadata)
importFrom(alabaster.base,acquireFile)
importFrom(alabaster.base,acquireMetadata)
importFrom(alabaster.base,addMissingPlaceholderAttributeForHdf5)
importFrom(alabaster.base,loadObject)
importFrom(alabaster.base,saveObject)
importFrom(alabaster.base,transformVectorForHdf5)
importFrom(rhdf5,H5Fclose)
importFrom(rhdf5,H5Fopen)
importFrom(rhdf5,H5Gclose)
importFrom(rhdf5,H5Gopen)
importFrom(rhdf5,H5Sunlimited)
importFrom(rhdf5,h5createDataset)
importFrom(rhdf5,h5createFile)
importFrom(rhdf5,h5createGroup)
importFrom(rhdf5,h5ls)
importFrom(rhdf5,h5read)
importFrom(rhdf5,h5readAttributes)
importFrom(rhdf5,h5write)
importFrom(rhdf5,h5writeAttribute)
importFrom(rhdf5,h5writeDataset)
useDynLib(alabaster.matrix, .registration=TRUE)
