#' DelayedArray wrapper seed
#'
#' Deprecated, use \link[DelayedArray]{DelayedUnaryIsoOp} instead.
#'
#' @aliases
#' WrapperArraySeed-class
#' WrapperArray-class
#' WrapperArraySeed
#' WrapperArray
#' loadWrapperArray
#'
#' @name WrapperArraySeed
NULL
