# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

lowest_double <- function() {
    .Call(`_alabaster_matrix_lowest_double`)
}

highest_double <- function() {
    .Call(`_alabaster_matrix_highest_double`)
}

collect_double_attributes <- function(x) {
    .Call(`_alabaster_matrix_collect_double_attributes`, x)
}

